# -*- Mode: Cython -*-

from cpython.bytes cimport PyBytes_FromStringAndSize
from libc.stdint cimport uint64_t, uint32_t, uint16_t, uint8_t
from libc.errno cimport errno, EAGAIN as EWOULDBLOCK

import sys
W = sys.stderr.write

class MODE:
    SERVER = S2N_SERVER
    CLIENT = S2N_CLIENT

class Error (Exception):
    pass

class Want (Exception):
    pass

class WantRead (Want):
    pass

class WantWrite (Want):
    pass

cdef raise_s2n_error():
    raise Error (s2n_strerror (s2n_errno, "EN"))

cdef check (int n):
    if n != 0:
        raise_s2n_error()

def init():
    check (s2n_init())

def cleanup():
    check (s2n_cleanup())

init()

cdef class Config:

    #cdef s2n_config * c

    def __init__ (self):
        self.c = s2n_config_new()
        if not self.c:
            raise_s2n_error()

    def __del__ (self):
        if self.c:
            check (s2n_config_free (self.c))

    def set_cipher_preferences (self, bytes version):
        check (s2n_config_set_cipher_preferences (self.c, version))

    def add_cert_chain_and_key (self, bytes chain_pem, bytes skey_pem):
        check (s2n_config_add_cert_chain_and_key (self.c, chain_pem, skey_pem))

    def add_cert_chain_and_key_with_status (self, bytes chain_pem, bytes skey_pem):
        cdef uint8_t status[512]
        check (s2n_config_add_cert_chain_and_key_with_status (self.c, chain_pem, skey_pem, &status[0], sizeof(status)))
        return <char*>status

    def add_dhparams (self, bytes dhparams_pem):
        check (s2n_config_add_dhparams (self.c, dhparams_pem))

    def set_protocol_preferences (self, protocols):
        cdef char * protos[50]
        cdef int count = 0
        assert (len(protocols) < 50)
        for i, proto in enumerate (protocols):
            protos[i] = proto
            count += 1
        check (s2n_config_set_protocol_preferences (self.c, <const char **> protos, count))

    def set_status_request_type (self, s2n_status_request_type stype):
        check (s2n_config_set_status_request_type (self.c, stype))

cdef maybe_string (char * s):
    if s:
        return s
    else:
        return None

cdef class Connection:

    #cdef s2n_connection * conn

    def __init__ (self, s2n_mode mode):
        self.conn = s2n_connection_new (mode)
        if not self.conn:
            raise_s2n_error()

    def __del__ (self):
        if self.conn:
            check (s2n_connection_free (self.conn))

    def set_config (self, Config cfg):
        check (s2n_connection_set_config (self.conn, cfg.c))

    def set_fd (self, int readfd):
        check (s2n_connection_set_fd (self.conn, readfd))

    def set_read_fd (self, int readfd):
        check (s2n_connection_set_read_fd (self.conn, readfd))

    def set_write_fd (self, int readfd):
        check (s2n_connection_set_write_fd (self.conn, readfd))

    def set_server_name (self, bytes server_name):
        check (s2n_set_server_name (self.conn, server_name))

    def get_server_name (self):
        return maybe_string (s2n_get_server_name (self.conn))

    def set_blinding (self, s2n_blinding blinding):
        check (s2n_connection_set_blinding (self.conn, blinding))

    def get_delay (self):
        return s2n_connection_get_delay (self.conn)

    def get_wire_bytes (self):
        return (
            s2n_connection_get_wire_bytes_in (self.conn),
            s2n_connection_get_wire_bytes_out (self.conn),
        )

    def get_client_hello_version (self):
        return s2n_connection_get_client_hello_version (self.conn)

    def get_client_protocol_version (self):
        return s2n_connection_get_client_protocol_version (self.conn)

    def get_server_protocol_version (self):
        return s2n_connection_get_server_protocol_version (self.conn)

    def get_actual_protocol_version (self):
        return s2n_connection_get_actual_protocol_version (self.conn)

    def get_application_protocol (self):
        return maybe_string (s2n_get_application_protocol (self.conn))

    def get_ocsp_response (self):
        cdef const uint8_t * r
        cdef uint32_t length
        r = s2n_connection_get_ocsp_response (self.conn, &length)
        return r[:length]

    def get_alert (self):
        return s2n_connection_get_alert (self.conn)

    def get_cipher (self):
        return maybe_string (s2n_connection_get_cipher (self.conn))

    # I/O

    # override these if you're using an event-driven system.
    cdef want_read (self):
        raise WantRead

    cdef want_write (self):
        raise WantWrite

    cdef bint _check_blocked (self, int r, s2n_blocked_status blocked, int errno) except -1:
        if r < 0:
            if errno == EWOULDBLOCK:
                if blocked == S2N_BLOCKED_ON_READ:
                    self.want_read()
                    return True
                elif blocked == S2N_BLOCKED_ON_WRITE:
                    self.want_write()
                    return True
                else:
                    raise_s2n_error()
            else:
                raise_s2n_error()
        else:
            return False

    cpdef negotiate (self):
        cdef s2n_blocked_status blocked = S2N_NOT_BLOCKED
        cdef int r = 0
        with nogil:
            r = s2n_negotiate (self.conn, &blocked)
        self._check_blocked (r, blocked, errno)
        return blocked

    cpdef send (self, bytes data, int pos=0):
        cdef s2n_blocked_status blocked = S2N_NOT_BLOCKED
        cdef ssize_t n
        cdef const char * p = <const char *>(data) + pos
        cdef ssize_t dlen = len(data)
        assert (pos < len(data))
        while 1:
            with nogil:
                n = s2n_send (self.conn, p, dlen - pos, &blocked)
            if not self._check_blocked (n, blocked, errno):
                break
        return n, blocked

    cpdef recv (self, ssize_t size):
        cdef s2n_blocked_status blocked = S2N_NOT_BLOCKED
        cdef bytes result = PyBytes_FromStringAndSize (NULL, size)
        cdef char * p = <char*>result
        cdef ssize_t n = 0
        cdef int saved_errno = 0
        while 1:
            with nogil:
                n = s2n_recv (self.conn, p, size, &blocked)
                saved_errno = errno
            if not self._check_blocked (n, blocked, saved_errno):
                break
        return result[:n], blocked

    cpdef shutdown (self):
        cdef s2n_blocked_status blocked = S2N_NOT_BLOCKED
        cdef int r = 0
        while 1:
            with nogil:
                r = s2n_shutdown (self.conn, &blocked)
            if not self._check_blocked (r, blocked, errno):
                break
        return blocked



# -*- Mode: Python -*-

import socket
import unittest

__version__ = '0.1.1'

from .cys2n import *

protocol_version_map = {
    'SSLv2' : 20,
    'SSLv3' : 30,
    'TLS10' : 31,
    'TLS11' : 32,
    'TLS12' : 33,
}

class PROTOCOL:
    reverse_map = {}

for name, val in protocol_version_map.items():
    setattr (PROTOCOL, name, val)
    PROTOCOL.reverse_map[val] = name

class s2n_socket:

    def __init__ (self, cfg, pysock, conn=None):
        self.cfg = cfg
        self.sock = pysock
        self.fd = pysock.fileno()
        self.conn = conn
        self.negotiated = False

    def __repr__ (self):
        return '<s2n sock=%r conn=%r @%x>' % (self.sock, self.conn, id (self))

    def bind (self, *args, **kwargs):
        return self.sock.bind (*args, **kwargs)

    def listen (self, *args, **kwargs):
        return self.sock.listen (*args, **kwargs)

    def accept (self):
        sock, addr = self.sock.accept()
        conn = Connection (MODE.SERVER)
        conn.set_config (self.cfg)
        conn.set_fd (sock.fileno())
        # XXX verify
        new = self.__class__ (self.cfg, sock, conn)
        return new, addr

    # XXX client mode as yet untested.
    def connect (self, addr):
        self.sock.connect (addr)
        self.conn = Connection (MODE.CLIENT)
        self.conn.set_config (self.cfg)
        self.conn.set_fd (self.fd)

    def _check_negotiated (self):
        if not self.negotiated:
            self.negotiate()

    def negotiate (self):
        while 1:
            blocked = self.conn.negotiate()
            if not blocked:
                break
        self.negotiated = True

    def recv (self, block_size):
        self._check_negotiated()
        r = []
        left = block_size
        while left:
            b, more = self.conn.recv (left)
            r.append (b)
            if not more:
                break
            else:
                left -= len(b)
        return b''.join (r)

    def send (self, data):
        self._check_negotiated()
        pos = 0
        left = len(data)
        while left:
            n, more = self.conn.send (data, pos)
            pos += n
            if not more:
                break
            else:
                pass
            left -= n
        return pos

    def shutdown (self, how=None):
        more = 1
        while more:
            more = self.conn.shutdown()

    def close (self):
        try:
            self.shutdown()
        finally:
            self.sock.close()




# -*- Mode: Cython -*-

from cpython.bytes cimport PyBytes_FromStringAndSize
from libc.stdint cimport uint64_t, uint32_t, uint16_t, uint8_t
from libc.errno cimport errno, EAGAIN as EWOULDBLOCK

cdef extern from "s2n.h" nogil:
    struct s2n_config
    int s2n_errno
    int s2n_init ()
    int s2n_cleanup ()

    s2n_config *s2n_config_new ()
    int s2n_config_free (s2n_config *config)
    int s2n_config_free_dhparams (s2n_config *config)
    int s2n_config_free_cert_chain_and_key (s2n_config *config)
    const char *s2n_strerror (int error, const char *lang)
    int s2n_config_add_cert_chain_and_key (
        s2n_config *config,
        char *cert_chain_pem,
        char *private_key_pem
    )
    int s2n_config_add_cert_chain_and_key_with_status (
        s2n_config *config,
        char *cert_chain_pem,
        char *private_key_pem,
        const uint8_t *status,
        uint32_t length
    )
    int s2n_config_add_dhparams (s2n_config *config, char *dhparams_pem)
    int s2n_config_set_key_exchange_preferences (s2n_config *config, const char *preferences)
    int s2n_config_set_cipher_preferences (
        s2n_config *config,
        const char *version
    )
    int s2n_config_set_protocol_preferences (
        s2n_config *config,
        const char * const *protocols,
        int protocol_count
    )

    ctypedef enum s2n_status_request_type:
        S2N_STATUS_REQUEST_NONE = 0,
        S2N_STATUS_REQUEST_OCSP = 1

    int s2n_config_set_status_request_type (s2n_config *config, s2n_status_request_type type)

    struct s2n_connection

    ctypedef enum s2n_mode:
        S2N_SERVER,
        S2N_CLIENT

    s2n_connection *s2n_connection_new (s2n_mode mode)
    int s2n_connection_set_config (s2n_connection *conn, s2n_config *config)

    int s2n_connection_set_fd (s2n_connection *conn, int readfd)
    int s2n_connection_set_read_fd (s2n_connection *conn, int readfd)
    int s2n_connection_set_write_fd (s2n_connection *conn, int readfd)

    ctypedef enum s2n_blinding:
        S2N_BUILT_IN_BLINDING,
        S2N_SELF_SERVICE_BLINDING

    int s2n_connection_set_blinding (s2n_connection *conn, s2n_blinding blinding)
    int s2n_connection_get_delay (s2n_connection *conn)

    int s2n_set_server_name (s2n_connection *conn, const char *server_name)
    const char *s2n_get_server_name (s2n_connection *conn)
    const char *s2n_get_application_protocol (s2n_connection *conn)
    const uint8_t *s2n_connection_get_ocsp_response (s2n_connection *conn, uint32_t *length)

    ctypedef enum s2n_blocked_status:
        S2N_NOT_BLOCKED = 0,
        S2N_BLOCKED_ON_READ,
        S2N_BLOCKED_ON_WRITE,

    int s2n_negotiate (s2n_connection *conn, s2n_blocked_status *blocked)
    ssize_t s2n_send (s2n_connection *conn, void *buf, ssize_t size, s2n_blocked_status *blocked)
    ssize_t s2n_recv (s2n_connection *conn,  void *buf, ssize_t size, s2n_blocked_status *blocked)

    int s2n_connection_wipe (s2n_connection *conn)
    int s2n_connection_free (s2n_connection *conn)
    int s2n_shutdown (s2n_connection *conn, s2n_blocked_status *blocked)

    uint64_t s2n_connection_get_wire_bytes_in (s2n_connection *conn)
    uint64_t s2n_connection_get_wire_bytes_out (s2n_connection *conn)
    int s2n_connection_get_client_protocol_version (s2n_connection *conn)
    int s2n_connection_get_server_protocol_version (s2n_connection *conn)
    int s2n_connection_get_actual_protocol_version (s2n_connection *conn)
    int s2n_connection_get_client_hello_version (s2n_connection *conn)
    const char *s2n_connection_get_cipher (s2n_connection *conn)
    int s2n_connection_get_alert (s2n_connection *conn)

cdef class Config:
    cdef s2n_config * c

cdef class Connection:
    cdef s2n_connection * conn
    cdef want_read (self)
    cdef want_write (self)
    cdef bint _check_blocked (self, int r, s2n_blocked_status blocked, int err) except -1
    cpdef negotiate (self)
    cpdef send (self, bytes data, int pos=?)
    cpdef recv (self, ssize_t size)
    cpdef shutdown (self)



# -*- Mode: Python -*-

from setuptools import setup, find_packages
from distutils.extension import Extension

from Cython.Build import cythonize

ext_modules = cythonize ([ Extension ('cys2n.cys2n', ['cys2n/cys2n.pyx'], libraries = ['s2n'])])

setup (
    name='cys2n',
    version='0.1.1',
    description='Cython wrapper for Amazon s2n',
    author='Sam Rushing',
    author_email='gitsam@rushing.nightmare.com',
    license="BSD",
    url="http://github.com/samrushing/cys2n",
    packages = find_packages(),
    package_data = {'cys2n': ['*.pxd']},
    ext_modules= ext_modules,
    download_url = 'https://pypi.python.org/pypi?name=cys2n',
)

